using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Reflection;

namespace Unum.Benefits.UI.Common.Binding
{
    /// <summary>
    /// A bound collection of BindingAdapter derived items.
    /// </summary>
    /// <typeparam name="TType">The type of BindingAdapter derived items in the collection.</typeparam>
    public class BindingAdapterCollection<TType> : List<TType>, ITypedList, IBindingAdapterCollection where TType : BindingAdapter
    {
        #region Declarations

        /// <summary>
        /// A collection of property descriptors for the items in this collection.
        /// </summary>
        private static PropertyDescriptorCollection propDescriptors = BindingAdapter.GetPropertyDescriptors<TType>();

        /// <summary>
        /// The collection's list name.
        /// </summary>
        private string _listName;

        #endregion

        #region Constructors

        /// <summary>
        /// Constructor.
        /// </summary>
        public BindingAdapterCollection()
        {
            _listName = "Binding adapter list";
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="listName">The collection's list name.</param>
        public BindingAdapterCollection(string listName)
        {
            _listName = listName;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Checks to see if an infinite loop is present in the given path of property descriptors.
        /// An ApplicationException will be thrown if an infinite loop is detected.
        /// </summary>
        /// <param name="listAccessors">The current path of property descriptors.</param>
        private void CheckForInfiniteLoop(PropertyDescriptor[] listAccessors)
        {
            if (listAccessors == null)
                return;

            for (int i = 0; i < listAccessors.Length; i++)
                for (int j = i + 1; j < listAccessors.Length; j++)
                    if (listAccessors[i].PropertyType == listAccessors[j].PropertyType)
                        throw new ApplicationException("An infinite loop was detected in the binding schema.");
        }

        #endregion

        #region ITypedList Members

        /// <summary>
        /// Describes the properties for the BindingAdapters in this collection.
        /// </summary>
        /// <param name="listAccessors">The path of property descriptors that leads to the one requested.</param>
        /// <returns>Property descriptors for the requested BindingAdapter.</returns>
        public PropertyDescriptorCollection GetItemProperties(PropertyDescriptor[] listAccessors)
        {
            CheckForInfiniteLoop(listAccessors);
            
            if (listAccessors == null)
                return propDescriptors;

            PropertyDescriptor propDesc = listAccessors[listAccessors.Length - 1];

            // process BindingAdapterCollection types
            if (typeof(IBindingAdapterCollection).IsAssignableFrom(propDesc.PropertyType))
            {
                Type[] genericTypes = propDesc.PropertyType.GetGenericArguments();
                if (genericTypes.Length == 1)
                {
                    Type adapterType = genericTypes[0];
                    if (typeof(BindingAdapter).IsAssignableFrom(adapterType))
                    {
                        return BindingAdapter.GetPropertyDescriptors(adapterType);
                    }
                }
            }

            return TypeDescriptor.GetProperties(propDesc.PropertyType);
        }

        /// <summary>
        /// The collection's list name.
        /// </summary>
        /// <param name="listAccessors">Multiple collection level names are not supported.</param>
        /// <returns>The collection's list name.</returns>
        public string GetListName(PropertyDescriptor[] listAccessors)
        {
            return _listName;
        }

        #endregion
    }
}
